#include "precomp.hpp"
#include "grfmt_webp.hpp"
    
#ifdef _MSC_VER
#if defined _M_X64
#pragma comment(lib, "libwebp.lib")
#else
#pragma comment(lib, "libwebp.lib")
#endif
#endif

#include "opencv2/imgproc/imgproc.hpp"

namespace cv
{
#include "webp/encode.h"
#include "webp/decode.h"

WebPDecoder::WebPDecoder()
{
    m_webp = NULL;
    m_src = NULL;
    m_size = 0;
    m_has_alpha = 0;
}


void WebPDecoder::close()
{
    if( m_webp )
    {
        FILE* webp = (FILE*)m_webp;
        fclose( webp );
        m_webp = NULL;
    }
    if( m_src )
    {
        free(m_src);
        m_src = NULL;
    }
}

WebPDecoder::~WebPDecoder()
{
    close();
}

ImageDecoder WebPDecoder::newDecoder() const
{
    return new WebPDecoder;
}

bool WebPDecoder::readHeader()
{
    bool result = false;

    close();
    FILE *webp = fopen(m_filename.c_str(), "rb" );
    WebPDecoderConfig config;

    if( webp )
    {
        fpos_t pos = {0};
        uint32_t size = 0;
        int ret = 0;
        
        fseek(webp, 0, SEEK_END); 
        fgetpos(webp, &pos); 
        fseek(webp, 0, SEEK_SET);

        m_webp = webp;
        
#if defined __linux || defined __linux__
        size = pos.__pos;
#else
        size = (uint32_t)pos;
#endif
        uint8_t* src = new uint8_t[(int)size];
        fread(src, sizeof(uint8_t), (size_t)size, webp);
        WebPInitDecoderConfig(&config);

        ret = WebPGetFeatures(src, (uint32_t)size, &config.input);
        if(ret ==  VP8_STATUS_OK)
        {
            m_width = config.input.width;
            m_height = config.input.height;
            m_has_alpha = config.input.has_alpha;
            if(config.input.has_alpha) m_type = CV_8UC4;
            else                       m_type = CV_8UC3;
            m_size = size;
            m_src = src;
            result = true;
        }
    }

    if( !result )
        close();

    return result;
}


bool  WebPDecoder::readData( Mat& img )
{
    bool result = false;
    int channels = img.channels();
    bool color = channels > 1;
    uchar* data = img.data;
    int step = (int)img.step;
    
    if( img.depth() != CV_8U )
        return false;
#if 0    
    if( img.depth() == CV_8U)
        m_type = CV_MAKETYPE(CV_8U, channels);
    else
        m_type = CV_MAKETYPE(CV_16U, channels);
#endif

    if( m_webp && m_width && m_height )
    {
        if(m_has_alpha && channels == 4)
        {
            uint8_t* buf = WebPDecodeBGRA((uint8_t*)m_src, (uint32_t)m_size, &m_width, &m_height);
            if(color)
            {
                memcpy(img.data, buf, m_width * m_height * channels);
            }
            else
            {
                for( int y = 0; y < m_height; y++, data += step, buf += (step * 4))
                {
                    icvCvt_BGRA2Gray_8u_C4C1R( buf, 0, data, 0, cvSize(m_width, 1));
                }
            }
        }
        else
        {
            uint8_t* buf = WebPDecodeBGR((uint8_t*)m_src, (uint32_t)m_size, &m_width, &m_height);
            if(color)
            {
                memcpy(img.data, buf, m_width * m_height * channels);
            }
            else
            {
                for( int y = 0; y < m_height; y++, data += step, buf += (step * 3))
                {
                    icvCvt_BGR2Gray_8u_C3C1R( buf, 0, data, 0, cvSize(m_width, 1));
                }
            }
        }
        result = true;
    }

    close();
    return result;
}


//////////////////////////////////////////////////////////////////////////////////////////

WebPEncoder::WebPEncoder()
{
    m_description = "WebP Files (*.webp)";
    m_buf_supported = true;
}

WebPEncoder::~WebPEncoder()
{
}

ImageEncoder WebPEncoder::newEncoder() const
{
    return new WebPEncoder;
}

bool  WebPEncoder::write( const Mat& img, const vector<int>& params )
{
    int channels = img.channels();
    int width = img.cols, height = img.rows;
    int depth = img.depth();
    const Mat* pimage = &img;
    Mat temp;
    int quality = 95;
    bool isLossless = false;
    bool hasAplha = false;
    size_t s = 0;

    if( depth != CV_8U)
    {
        return false;
    }

    for( size_t i = 0; i < params.size(); i += 2 )
    {
        if( params[i] == CV_IMWRITE_WEBP_QUALITY )
        {
            quality = params[i+1];
            quality = MIN(MAX(quality, 0), 100);
        }
        else if( params[i] == CV_IMWRITE_WEBP_LOSSLESS )
        {
            isLossless = params[i+1] != 0;
        }
    }

    if( channels == 1 )
    {
        cvtColor(*pimage, temp, CV_GRAY2BGR);
        pimage = &temp;
        channels = 3;
    }
    else if( channels == 4 )
    {
        hasAplha = true;
    }

    uint8_t* out = new uint8_t[width * height * channels];

    if(isLossless)
    {
        if(hasAplha) s = WebPEncodeLosslessBGRA(pimage->data, width, height, width * channels, &out);
        else         s = WebPEncodeLosslessBGR(pimage->data, width, height, width * channels, &out);
    }
    else
    {
        if(hasAplha) s = WebPEncodeBGRA(pimage->data, width, height, width * channels, (float)quality, &out);
        else         s = WebPEncodeBGR(pimage->data, width, height, width * channels, (float)quality, &out);
    }

    FILE* fp = fopen(m_filename.c_str(), "wb");
    fwrite(out, s, sizeof(uint8_t), fp);
    fclose(fp);
    delete[] out;

    return true;
}

}
